"""
nsarchive - API-wrapper pour récupérer des données liées à Nation.

Version: 3.0.0a3
License: GPL-3.0
Auteur : happex <110610727+okayhappex@users.noreply.github.com>

Dependencies:
- Python ^3.10
- pillow ^10.4

Le fichier README.md fournit des détails supplémentaires pour l'utilisation.
"""

# Import des types 
from .cls.base import NSID
from .cls.archives import *
from .cls.entities import *
from .cls.republic import *
from .cls.economy import *

# Import des instances
from .instances._economy import EconomyInstance
from .instances._entities import EntityInstance
from .instances._republic import RepublicInstance


from .base import NSID

class BankAccount:
    """
    Compte en banque d'une entité, individuelle ou collective.

    ## Attributs
    - id: `NSID`\n
        Identifiant du compte
    - owner: `NSID`\n
        Identifiant du titulaire du compte
    - amount: `int`\n
        Somme d'argent totale sur le compte
    - frozen: `bool`\n
        État gelé ou non du compte
    - bank: `NSID`\n
        Identifiant de la banque qui détient le compte
    - income: `int`\n
        Somme entrante sur le compte depuis la dernière réinitialisation (tous les ~ 28 jours)
    """

    def __init__(self, id: NSID) -> None:
        self.id: NSID = NSID(id)
        self.owner: NSID = NSID(0)
        self.amount: int = 0
        self.frozen: bool = False
        self.bank: NSID = NSID("6")

        self.income: int = 0

class Item:
    """
    Article d'inventaire qui peut circuler sur le serveur

    ## Attributs
    - id: `NSID`\n
        Identifiant de l'objet
    - title: `str`\n
        Nom de l'objet
    - emoji: `str`\n
        Emoji lié à l'objet
    """

    def __init__(self, id: NSID) -> None:
        self.id: NSID = NSID(id)
        self.title: str = "Unknown Object"
        self.emoji: str = ":light_bulb:"

class Inventory:
    """
    Inventaire d'un membre

    ## Attributs
    - owner_id: `NSID`\n
        ID du propriétaire de l'inventaire
    - objects: `dict[str, NSID]`\n
        Collection d'objets et leur quantité
    """

    def __init__(self, owner_id: NSID) -> None:
        self.owner_id: NSID = NSID(owner_id)
        self.objects: dict[str, NSID] = {}

    def append(self, item: Item, quantity: int = 1):
        if item.id in self.objects.keys():
            self.objects[item.id] += quantity
        else:
            self.objects[item.id] = quantity

    def throw(self, item: Item, quantity: int = 1):
        if item.id in self.objects.keys():
            if self.objects[item.id] > quantity:
                self.objects[item.id] -= quantity
            else:
                self.objects[item.id] = 0

class Sale:
    """
    Vente mettant en jeu un objet

    ## Attributs
    - id: `NSID`\n
        Identifiant de la vente
    - item: `NSID`\n
        Identifiant de l'objet mis en vente
    - quantity: `int`\n
        Quantité d'objets mis en vente
    - price: `int`\n
        Prix total du lot
    - seller_id: `NSID`\n
        Identifiant du vendeur
    """

    def __init__(self, id: NSID, item: Item) -> None:
        self.id: NSID = NSID(id)
        self.item: NSID = NSID(item.id)
        self.quantity: int = 1

        self.price: int = 0
        self.seller_id: NSID = NSID('0')


import time

from ..cls.base import *
from ..cls.archives import *
from ..cls.economy import *

class EconomyInstance(Instance):
    """Indisponible dans cette version."""

    def __init__(self, url: str, token: str) -> None:
        super().__init__(url, token)

    """
    ---- COMPTES EN BANQUE ----
    """

    def get_account(self, id: NSID) -> BankAccount:
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Récupère les informations d'un compte bancaire.

        ## Paramètres
        id: `NSID`\n
            ID du compte.

        ## Renvoie
        - `.BankAccount`
        """

        return BankAccount(NSID(id)) # Provisoire

        id = NSID(id)
        _data = self._get_by_ID('accounts', id)

        if _data is None:
            return None

        account = BankAccount(id)
        account.amount = _data['amount']
        account.frozen = _data['frozen']
        account.owner = NSID(_data['owner_id'])
        account.bank = _data['bank']
        account.income = _data['income']

        return account

    def save_account(self, account: BankAccount):
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Sauvegarde un compte bancaire dans la base de données.

        ## Paramètres
        - account: `.BankAccount`\n
            Compte à sauvegarder
        """

        return # Provisoire

        _data = {
            'id': NSID(account.id),
            'amount': account.amount,
            'frozen': account.frozen, 
            'owner_id': account.owner, 
            'bank': account.bank,
            'income': account.income
        }

        self._put_in_db('accounts', _data)

    def freeze_account(self, account: BankAccount):
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Gèle un compte bancaire pour empêcher toute transaction.

        ## Paramètres
        - account: `.BankAccount`\n
            Compte à geler
        """

        return # Provisoire

        account.id = NSID(account.id)
        account.frozen = True

        self.save_account(account)

    """
    ---- OBJETS & VENTES ----
    """

    def save_item(self, item: Item):
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Sauvegarde des infos à propos d'un item.

        ## Paramètres
        item: `.Item`\n
            Article à sauvegarder
        """

        return # Provisoire

        _item = item.__dict__
        self._put_in_db('items', _item)

    def get_item(self, id: NSID) -> Item | None:
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Récupère des informations à propos d'un item.

        ## Paramètres
        id: `NSID`\n
            ID de l'item

        ## Retourne
        - `.Item` si quelque chose est trouvé, sinon
        - `None`
        """

        return Item(NSID(id)) # Provisoire

        _item = self._get_by_ID('items', id)

        if _item is None:
            return

        item = Item(id)
        item.title = _item['title']
        item.emoji = _item['emoji']

        return item

    def delete_item(self, item: Item):
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Annule le référencement d'un item.

        ## Paramètres
        item: `.Item`\n
            Item à supprimer
        """

        return # Provisoire

        self._delete_by_ID('items', item.id)

    def get_sale(self, id: NSID) -> Sale | None:
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Récupère une vente disponible sur le marketplace.

        ## Paramètres
        id: `NSID`\n
            ID de la vente.

        ## Renvoie
        - `.Sale | None`: Le résultat de la vente
        """

        return Sale(NSID(id), Item(NSID(id))) # Provisoire

        id = NSID(id)

        _data = self._get_by_ID('market', id)

        if _data is None:
            return None

        item = self.get_item(_data['id'])

        sale = Sale(NSID(id), Item(_data['id']) if item is None else item)
        sale.__dict__ = _data

        return sale

    def sell_item(self, item: Item, quantity: int, price: int, seller: NSID):
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Vend un item sur le marché.

        ## Paramètres
        item: `.Item`\n
            Item à vendre
        quantity: `int`\n
            Nombre d'items à vendre
        price: `int`\n
            Prix à l'unité de chaque objet
        seller: `NSID`\n
            ID de l'auteur de la vente
        """

        return # Provisoire

        sale = Sale(NSID(round(time.time()) * 16 ** 3), item)
        sale.quantity = quantity
        sale.price = price
        sale.seller_id = seller

        _data = sale.__dict__.copy()

        self._put_in_db('market', _data)

    def delete_sale(self, sale: Sale) -> None:
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Annule une vente sur le marketplace.
        """

        return # Provisoire

        sale.id = NSID(sale.id)
        self._delete_by_ID('market', NSID(sale.id))

    """
    ---- INVENTAIRES ----
    """

    def get_inventory(self, id: NSID) -> Inventory | None:
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Récupérer un inventaire dans la base des inventaires.

        ## Paramètres
        id: `NSID`\n
            ID du propriétaire de l'inventaire

        ## Retourne
        - `.Inventory | None`: L'inventaire s'il a été trouvé
        """

        return Inventory(NSID(id)) # Provisoire

        _data = self._get_by_ID('inventories', id)

        if _data is None:
            return None

        inventory = Inventory(id)

        for _item in _data['objects']:
            item = self.get_item(_item)

            inventory.objects.append(item)

        return inventory

    def save_inventory(self, inventory: Inventory):
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Sauvegarder un inventaire

        ## Paramètres
        inventory: `.Inventory`\n
            Inventaire à sauvegarder
        """

        return # Provisoire

        _data = inventory.__dict__

        self._put_in_db('inventories', _data)

    def delete_inventory(self, inventory: Inventory):
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Supprime un inventaire

        ## Paramètres
        inventory: `.Inventory`
            Inventaire à supprimer
        """

        return # Provisoire

        self._delete_by_ID('inventories', inventory.owner_id)

    """
    ---- ARCHIVES ----
    """

    def _add_archive(self, archive: Archive):
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Ajoute une archive d'une transaction ou d'une vente dans la base de données.

        ## Paramètres
        - archive: `.Archive`\n
            Archive à ajouter
        """

        return # Provisoire

        archive.id = NSID(archive.id)
        archive.author = NSID(archive.author)
        archive.target = NSID(archive.target)

        _data = archive.__dict__.copy()

        if type(archive) == Transaction:
            _data['_type'] = "transaction"
        else:
            _data['_type'] = "action"

        self._put_in_db('archives', _data)

    def _get_archive(self, id: NSID) -> Archive | Transaction:
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Récupère une archive spécifique.

        ## Paramètres
        id: `NSID`\n
            ID de l'archive.

        ## Renvoie
        - `.Archive | .Transaction`
        """

        return Archive() # Provisoire

        id = NSID(id)
        _data = self._get_by_ID('archives', id)

        if _data is None:
            return None

        if _data['_type'] == "transaction":
            archive = Transaction(_data['author'], _data['target'])
        else:
            archive = Archive(_data['author'], _data['target'])

        archive.id = id
        archive.action = _data['action']
        archive.date = _data['date']
        archive.details = _data['details']

        return archive

    def _fetch_archives(self, **query) -> list[ Archive | Transaction ]:
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Récupère une liste d'archives correspondant à la requête.

        ## Paramètres
        query: `dict`\n
            Requête pour filtrer les archives.

        ## Renvoie
        - `list[.Archive | .Transaction]`
        """

        return [] # Provisoire

        _res = self.fetch('archives', **query)

        return [ self._get_archive(archive['id']) for archive in _res ]


import io
import json
import requests
import typing
import warnings

class NSID(str):
    """
    Nation Server ID

    ID unique et universel pour l'ensemble des entités et évènements. Il prend les `int`, les `str` et les autres instances `NSID` pour les convertir en un identifiant hexadécimal.
    """
    unknown = "0"
    admin = "1"
    gov = "2"
    court = "3"
    assembly = "4"
    office = "5"
    hexabank = "6"
    archives = "7"

    maintenance_com = "101"
    audiovisual_dept = "102"
    interior_dept = "103"
    justice_dept = "104"
    egalitary_com = "105"
    antifraud_dept = "106"

    def __new__(cls, value):
        if type(value) == int:
            value = hex(value)
        elif type(value) in (str, NSID):
            value = hex(int(value, 16))
        else:
            raise TypeError(f"<{value}> is not NSID serializable")

        if value.startswith("0x"):
            value = value[2:]

        instance = super(NSID, cls).__new__(cls, value.upper())
        return instance

class Instance:
    """
    Instance qui servira de base à toutes les instances.
    """

    def __init__(self, url: str, token: str = None):
        self.url = url
        self.token = token

        self.default_headers = {
            "Authorization": f"Bearer {self.token}",
            "Content-Type": "application/json",
        }

    def request_token(self, username: str, password: str) -> str | None:
        res = requests.post(f"{self.url}/auth/login", json = {
            "username": username,
            "password": password
        })

        if res.status_code == 200:
            return res.json()["token"]
        elif res.status_code in (401, 403):
            raise PermissionError(res.json()['message'])
        else:
            raise Exception(f"Error {res.status_code}: {res.json()['message']}")

    def _get_item(self, endpoint: str, body: dict = None, headers: dict = None) -> dict:
        """
        Récupère des données JSON depuis l'API

        ## Paramètres
        endpoint: `str`:
            Endpoint de l'URL
        headers: `dict` (optional)
            Headers à envoyer
        body: `dict` (optional)
            Données à envoyer

        ## Renvoie
        - `list` de tous les élements correspondants
        - `None` si aucune donnée n'est trouvée
        """

        if not headers:
            headers = self.default_headers

        res = requests.get(f"{self.url}/{endpoint}", headers = headers, json = body, timeout = 5)

        if 200 <= res.status_code < 300:
            return res.json()
        elif res.status_code == 404:
            return
        elif res.status_code in (403, 401):
            raise PermissionError(res.json()['message'])
        else:
            raise Exception(f"Error {res.status_code}: {res.json()['message']}")

    def _get_by_ID(self, _class: str, id: NSID) -> dict:
        _data = self._get_item(f"/model/{_class}/{id}")

        return _data

    def _put_in_db(self, endpoint: str, body: dict, headers: dict = None, use_PUT: bool = False) -> None:
        """
        Publie des données JSON dans une table nation-db.

        ## Paramètres
        endpoint: `str`
            Endpoint de l'URL
        body: `dict`
            Données à envoyer
        headers: `dict` (optionnel)
            Headers à envoyer
        """

        if not headers:
            headers = headers

        if use_PUT:
            res = requests.put(f"{self.url}/{endpoint}", headers = headers, json = body)
        else:
            res = requests.post(f"{self.url}/{endpoint}", headers = headers, json = body)

        if 200 <= res.status_code < 300:
            return res.json()
        else:
            print(res.text)
            res.raise_for_status()

    def _delete(self, _class: str, ids: list[NSID]) -> None:
        """
        Supprime des données JSON dans une table nation-db.

        ## Paramètres
        _class: `str`
            Classe des entités à supprimer
        ids: `list[NSID]`
            ID des entités à supprimer
        """

        res = requests.post(f"{self.url}/delete_{_class}", json = { "ids": ids })

        if 200 <= res.status_code < 300:
            return res.json()
        elif res.status_code in (403, 401):
            raise PermissionError(res.json()['message'])
        else:
            raise Exception(f"Error {res.status_code}: {res.json()['message']}")

    def _delete_by_ID(self, _class: str, id: NSID):
        warnings.showwarning("Method '_delete_by_id' is deprecated. Use '_delete' instead.")
        self._delete(_class, id)

    def fetch(self, _class: str, **query: typing.Any) -> list:
        res = requests.get(f"{self.url}/fetch/{_class}", params = query)

        if res.status_code == 200:
            matches = res.json()
        elif res.status_code in (401, 403):
            matches = []
        else:
            res.raise_for_status()

        return matches


    def _upload_file(self, bucket: str, name: str, data: bytes, overwrite: bool = False, headers: dict = None) -> dict:
        """
        Envoie un fichier dans un bucket nation-db.

        ## Paramètres
        bucket: `str`
            Nom du bucket où le fichier sera stocké
        name: `str`
            Nom du fichier dans le drive
        data: `bytes`
            Données à uploader
        overwrite: `bool` (optional)
            Overwrite ou non
        headers: `dict` (optional)
            Headers à envoyer

        ## Renvoie
        - `dict` contenant les informations de l'upload si réussi
        - `None` en cas d'échec
        """

        if not headers:
            headers = self.default_headers
            headers['Content-Type'] = 'image/png'

        body = {
            "name": name,
            "overwrite": json.dumps(overwrite)
        }

        file = ("file", "image/png", data)

        res = requests.put(f"{self.url}/upload_file/{bucket}", headers = headers, json = body, files = [ file ])

        if res.status_code == 200:
            return res.json()
        elif res.status_code in (403, 401):
            raise PermissionError(res.json()['message'])
        elif res.status_code == 409:
            raise FileExistsError(res.json()['message'])
        else:
            raise Exception(f"Error {res.status_code}: {res.json()['message']}") 

    def _download_from_storage(self, bucket: str, path: str, headers: dict = None) -> bytes:
        """
        Télécharge un fichier depuis le stockage nation-db.

        ## Paramètres
        bucket: `str`\n
            Nom du bucket où il faut chercher le fichier 
        path: `str`\n
            Chemin du fichier dans le bucket

        ## Renvoie
        - Le fichier demandé en `bytes`
        """

        if not headers:
            headers = self.default_headers

        res = requests.get(f"{self.url}/drive/{bucket}/{path}", headers = headers)

        if res.status_code == 200:
            return res.json()
        elif res.status_code in (403, 401):
            raise PermissionError(res.json()['message'])
        else:
            raise Exception(f"Error {res.status_code}: {res.json()['message']}") 



import time

from ..cls.base import *
from ..cls.archives import *
from ..cls.republic import *


class RepublicInstance(Instance):
    """
        *INDISPONIBLE DANS CETTE VERSION.*\n
    Gère les interactions avec les votes, les archives de la république, et les fonctionnaires.

    ## Informations
    - Résultats des votes et procès: `.Vote | .Referendum | .Lawsuit`
    - Différentes institutions: `.State | .Administration | .Government | .Assembly | .Court | .PoliceForces`
    - Occupants des différents rôles et historique de leurs actions: `.Official`
    """

    def __init__(self, url: str, token: str) -> None:
        super().__init__(url, token)
    
    """
    ---- VOTES & REFERENDUMS ----
    """

    def get_vote(self, id: NSID) -> Vote | Referendum | Lawsuit:
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Récupère un vote spécifique.

        ## Paramètres
        id: `NSID`\n
            ID du vote.

        ## Renvoie
        - `.Vote | .Referendum | .Lawsuit`
        """

        return Vote(NSID(id), "Vote Inconnu") # Provisoire

        id = NSID(id)
        _data = self._get_by_ID('votes', id)

        if not _data: # Pas dans les votes -> peut-être dans les procès
            _data = self._get_by_ID('lawsuits', id)

        if not _data: # Le vote n'existe juste pas
            return None
        elif '_type' not in _data.keys(): # Le vote est un procès
            _data['_type'] = "lawsuit"

        if _data['_type'] == 'vote':
            vote = Vote(id, _data['title'])
        elif _data['_type'] == 'referendum':
            vote = Referendum(id, _data['title'])
            vote.choices = []
        elif _data['_type'] == 'lawsuit':
            vote = Lawsuit(id, _data['title'])
            vote.choices = []
        else:
            vote = Vote('0', 'Unknown Vote')

        vote.author = _data['author_id']
        vote.startDate = _data['start_date']
        vote.endDate = _data['end_date']

        for opt in _data['choices']:
            option = VoteOption(opt["id"], opt["title"])
            option.count = opt["count"]

            vote.choices.append(option)

        return vote

    def save_vote(self, vote: Vote | Referendum | Lawsuit):
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Sauvegarde un vote dans la base de données.

        ## Paramètres
        - vote: `.Vote`\n
            Vote à sauvegarder
        """

        return # Provisoire

        vote.id = NSID(vote.id)

        _data = {
            '_type':'vote' if type(vote) == Vote else\
                    'referendum' if type(vote) == Referendum else\
                    'lawsuit' if type(vote) == Lawsuit else\
                    'unknown',
            'id': NSID(vote.id),
            'title': vote.title,
            'author_id': NSID(vote.author),
            'start_date': vote.startDate,
            'end_date': vote.endDate,
            'choices': [ opt.__dict__ for opt in vote.choices ]
        }

        if type(vote) == Lawsuit:
            del _data['_type']
            self._put_in_db('lawsuits', _data)
        else:
            self._put_in_db('votes', _data)

    # Aucune possibilité de supprimer un vote

    """
    ---- INSTITUTION & MANDAT ----
    """

    def get_official(self, id: NSID, current_mandate: bool = True) -> Official:
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Récupère les informations d'un fonctionnaire (mandats, contributions).

        ## Paramètres
        id: `NSID`\n
            ID du fonctionnaire.
        current_mandate: `bool`\n
            Indique si l'on doit récupérer le mandat actuel ou les anciens mandats.

        ## Renvoie
        - `.Official`
        """

        return Official(NSID(id)) # Provisoire

        id = NSID(id)

        base = 'mandate' if current_mandate else 'archives'

        _contributions = self.fetch(base, author = id, _type = 'contrib')
        _mandates = self.fetch(base, target = id, _type = 'election') +\
                    self.fetch(base, target = id, _type = 'promotion')

        user = Official(id)
        for mandate in _mandates:
            if mandate['details']['position'].startswith('MIN'):
                mandate['details']['position'] = 'MIN'

            try:
                user.mandates[mandate['details']['position']] += 1
            except KeyError:
                user.mandates[mandate['details']['position']] = 1

        for contrib in _contributions:
            try:
                user.contributions[contrib['action']] += 1
            except KeyError:
                user.contributions[contrib['action']] = 1

        return user

    def get_institutions(self) -> State:
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Récupère l'état actuel des institutions de la république.
        """

        return State()

        admin = Administration()
        gov = Government(Official('0'))
        assembly = Assembly()
        court = Court()
        police_forces = PoliceForces()

        _get_position: list[dict] = lambda pos : self._select_from_db('functions', 'id', pos)[0]['users']

        admin.members = [ self.get_official(user) for user in _get_position('ADMIN') ]
        admin.president = self.get_official(0xF7DB60DD1C4300A) # happex (remplace Kheops pour l'instant)

        gov.president = self.get_official(0x0)

        minister = lambda code : self.get_official(_get_position(f'MIN_{code}')[0])
        gov.prime_minister = minister('PRIM')
        gov.economy_minister = minister('ECO')
        gov.inner_minister = minister('INN')
        gov.press_minister = minister('AUD')
        gov.justice_minister = minister('JUS')
        gov.outer_minister = minister('OUT')

        assembly.president = self.get_official(_get_position('PRE_AS')[0])
        assembly.members = [ self.get_official(user) for user in _get_position('REPR') ]

        court.president = gov.justice_minister
        court.members = [ self.get_official(user) for user in _get_position('JUDGE') ]

        police_forces.president = gov.inner_minister
        police_forces.members = [ self.get_official(user) for user in _get_position('POLICE') ]

        instits = State()
        instits.administration = admin
        instits.government = gov
        instits.court = court
        instits.assembly = assembly
        instits.police = police_forces

        return instits

    def update_institutions(self, institutions: State):
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Fonction communément appelée après un vote législatif ou une nomination.\n
        Celle-ci met à jour: Le gouvernement (président, ministres), les différents députés et leur président, les différents juges, les différents policiers.\n

        ## Paramètres
        institutions: `.State`\n
            Le nouvel état des institutions, à sauvegarder.
        """

        return # Provisoire

        get_ids = lambda institution : [ member.id for member in institutions.__getattribute__(institution).members ]

        self._put_in_db('functions', { 'id': 'ADMIN', 'users': get_ids('administration') })
        self._put_in_db('functions', { 'id': 'REPR', 'users': get_ids('assembly') })
        self._put_in_db('functions', { 'id': 'JUDGE', 'users': get_ids('court') })
        self._put_in_db('functions', { 'id': 'POLICE', 'users': get_ids('police') })

        self._put_in_db('functions', { 'id': 'PRE_AS', 'users': [ institutions.assembly.president.id ] })
        self._put_in_db('functions', { 'id': 'PRE_REP', 'users': [ institutions.government.president.id ] })

        self._put_in_db('functions', { 'id': 'MIN_PRIM', 'users': [ institutions.government.prime_minister.id ] })
        self._put_in_db('functions', { 'id': 'MIN_INN', 'users': [ institutions.government.inner_minister.id ] })
        self._put_in_db('functions', { 'id': 'MIN_JUS', 'users': [ institutions.government.justice_minister.id ] })
        self._put_in_db('functions', { 'id': 'MIN_ECO', 'users': [ institutions.government.economy_minister.id ] })
        self._put_in_db('functions', { 'id': 'MIN_AUD', 'users': [ institutions.government.press_minister.id ] })
        self._put_in_db('functions', { 'id': 'MIN_OUT', 'users': [ institutions.government.outer_minister.id ] })

    def new_mandate(self, institutions: State, weeks: int = 4):
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Fonction qui amène à supprimer toutes les archives du mandat précédent

        ## Paramètres
        - institutions: `.State`\n
            Nouvel État à sauvegarder
        - weeks: `int`\n
            Nombre de semaines du mandat
        """

        return # Provisoire

        for item in self.fetch('mandate'):
            if item['date'] >= round(time.time()) - weeks * 604800: # On évite de supprimer les informations écrites lors de la période définie
                self._delete_by_ID('mandate', item['id'])

        self.update_institutions(institutions)

    """
    ---- ARCHIVES ----
    """

    def _add_archive(self, archive: Archive):
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Ajoute une archive d'une action (élection, promotion, ou rétrogradation) dans la base de données.

        ## Paramètres
        - archive: `.Archive`\n
            Archive à ajouter
        """

        return # Provisoire

        archive.id = NSID(archive.id)
        _data = archive.__dict__.copy()

        if type(archive) == Election:
            _data['_type'] = "election"
        elif type(archive) == Promotion:
            _data['_type'] = "promotion"
        elif type(archive) == Demotion:
            _data['_type'] = "demotion"
        else:
            _data['_type'] = "unknown"

        self._put_in_db('archives', _data)
        self._put_in_db('mandate', _data) # Ajouter les archives à celle du mandat actuel

    def _get_archive(self, id: NSID) -> Archive | Election | Promotion | Demotion:
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Récupère une archive spécifique.

        ## Paramètres
        id: `NSID`\n
            ID de l'archive.

        ## Renvoie
        - `.Archive | .Election | .Promotion | .Demotion`
        """

        return Archive(NSID(id)) # Provisoire

        id = NSID(id)
        _data = self._get_by_ID('archives', id)

        if _data is None:
            return None

        if _data['_type'] == "election":
            archive = Election(_data['author'], _data['target'], _data['details']['position'])
        elif _data['_type'] == "promotion":
            archive = Promotion(_data['author'], _data['target'], _data['details']['position'])
        elif _data['_type'] == "demotion":
            archive = Demotion(_data['author'], _data['target'])
        else:
            archive = Archive(_data['author'], _data['target'])

        archive.id = id
        archive.action = _data['action']
        archive.date = _data['date']
        archive.details = _data['details']

        return archive

    def _fetch_archives(self, **query) -> list[ Archive | Election | Promotion | Demotion ]:
        """
        *INDISPONIBLE DANS CETTE VERSION.*\n
        Récupère une liste d'archives correspondant à la requête.

        ## Paramètres
        query: `dict`\n
            Requête pour filtrer les archives.

        ## Renvoie
        - `list[.Archive | .Election | .Promotion | .Demotion]`
        """

        return [] # Provisoire

        _res = self.fetch('archives', **query)
        
        return [ self._get_archive(archive['id']) for archive in _res ]



from nsarchive.cls.base import NSID

# Votes

class VoteOption:
    """
    Option disponible lors d'un vote

    ## Attributs
    - id: `str`\n
        Identifiant de l'option
    - title: `str`\n
        Label de l'option
    - count: `int`\n
        Nombre de sympathisants pour cette option
    """

    def __init__(self, id: str, title: str = None, count: int = 0):
        self.id = id
        self.title = title if title else id
        self.count = count

class Vote:
    """
    Classe de référence pour les différents votes du serveur

    ## Attributs
    - id: `NSID`\n
        Identifiant du vote
    - title: `str`\n
        Titre du vote
    - choices: list[.VoteOption]\n
        Liste des choix disponibles
    - author: `NSID`\n
        Identifiant de l'auteur du vote
    - startDate: `int`\n
        Date de début du vote
    - endDate: `int`\n
        Date limite pour voter
    """

    def __init__(self, id: NSID, title: str) -> None:
        self.id: NSID = NSID(id)
        self.title: str = title
        self.choices: list[VoteOption] = []
        self.author: NSID = NSID("0")
        self.startDate: int = 0
        self.endDate: int = 0

    def by_id(self, id: str) -> VoteOption:
        for opt in self.choices:
            if opt.id == id:
                return opt

    def sorted(self, titles_only: bool = False) -> list[VoteOption] | list[str]:
        sorted_list: list[VoteOption] = sorted(self.choices, lambda opt : opt.count)

        if titles_only:
            return [ opt.id for opt in sorted_list ]
        else:
            return sorted_list

class Referendum(Vote):
    """
    Vote à trois positions
    """

    def __init__(self, id: NSID, title: str) -> None:
        super().__init__(id, title)

        self.choices = [
            VoteOption('yes', 'Oui'),
            VoteOption('no', 'Non'),
            VoteOption('blank', 'Pas d\'avis'),
        ]

class Lawsuit(Vote):
    """
    Vote à trois positions pour un procès
    """

    def __init__(self, id: NSID, title: str) -> None:
        super().__init__(id, title)

        self.choices = [
            VoteOption('guilty', 'Coupable'),
            VoteOption('innocent', 'Innocent'),
            VoteOption('blank', 'Pas d\'avis'),
        ]


# Institutions (defs)

class Official:
    def __init__(self, id: NSID) -> None:
        self.id: NSID = NSID(id)

        self.mandates: int = {
            'PRE_REP': 0, # Président de la République
            'MIN': 0, # Différents ministres
            'PRE_AS': 0, # Président de l'Assemblée Nationale
            'REPR': 0 # Député
        }

        self.contributions: dict = {
            'propose_project': 0,
            'success_project': 0, 
            'vote_law': 0
        }

class Administration:
    def __init__(self) -> None:
        self.president: Official
        self.members: list[Official] = []

class Government:
    def __init__(self, president: Official) -> None:
        self.president: Official = president
        self.prime_minister: Official

        self.inner_minister: Official
        self.economy_minister: Official
        self.justice_minister: Official
        self.press_minister: Official
        self.outer_minister: Official

class Court:
    def __init__(self) -> None:
        self.president: Official
        # On discutera de la mise en place d'un potentiel président. Pour l'instant c'est le Ministre de la Justice
        self.members: list[Official] = []

class Assembly:
    def __init__(self) -> None:
        self.president: Official
        self.members: list[Official] = []

class PoliceForces:
    def __init__(self) -> None:
        self.president: Official
        self.members: list[Official] = []

class State:
    def __init__(self) -> None:
        self.administration: Administration
        self.government: Government
        self.court: Court
        self.assembly: Assembly
        self.police: PoliceForces


import requests
import time
import typing
import urllib
import warnings

from .base import NSID

from .. import utils

default_headers = {}

class Permission:
    def __init__(self, initial: str = "----"):
        self.append: bool
        self.manage: bool
        self.edit: bool
        self.read: bool

        self.load(initial)

    def load(self, val: str) -> None:
        if 'a' in val: self.append = True
        if 'm' in val: self.manage = True
        if 'e' in val: self.edit = True
        if 'r' in val: self.read = True

class PositionPermissions:
    """
    Permissions d'une position à l'échelle du serveur. Certaines sont attribuées selon l'appartenance à divers groupes ayant une position précise
    """

    def __init__(self) -> None:
        self.bank_accounts = Permission("a---") # APPEND = ouvrir un ou plusieurs comptes, MANAGE = voir les infos globales concernant les comptes en banque, EDIT = gérer des comptes en banque, READ = voir les infos d'un compte en banque individuel
        self.bots = Permission() # APPEND = publier un message sous l'identité d'un bot, MANAGE = proposer d'héberger un bot, EDIT = changer les paramètres d'un bot, READ = /
        self.constitution = Permission() # APPEND = laws.append, MANAGE = laws.manage, EDIT = modifier la constitution, READ = /
        self.database = Permission() # APPEND = créer des sous-bases de données, MANAGE = gérer la abse de données, EDIT = modifier les éléments, READ = avoir accès à toutes les données sans exception
        self.items = Permission("---r") # APPEND = vendre, MANAGE = gérer des items dont on n'est pas propriétaire (hors marketplace), EDIT = gérer des items dont on n'est pas propriétaire (dans le marketplace), READ = accéder au marketplace
        self.laws = Permission() # APPEND = proposer un texte de loi, MANAGE = accepter ou refuser une proposition, EDIT = modifier un texte, READ = /
        self.members = Permission("---r") # APPEND = créer des entités, MANAGE = modérer des entités (hors Discord), EDIT = modifier des entités, READ = voir le profil des entités
        self.money = Permission("---r") # APPEND = créer des entités, MANAGE = modérer des entités (hors Discord), EDIT = modifier des entités, READ = voir le profil des entités
        self.national_channel = Permission() # APPEND = prendre la parole sur la chaîne nationale, MANAGE = voir qui peut prendre la parole, EDIT = modifier le planning de la chaîne nationale, READ = /
        self.organizations = Permission("---r") # APPEND = créer une nouvelle organisation, MANAGE = exécuter des actions administratives sur les organisations, EDIT = modifier des organisations, READ = voir le profil de n'importe quelle organisation
        self.reports = Permission() # APPEND = déposer plainte, MANAGE = accépter ou refuser une plainte, EDIT = /, READ = accéder à des infos supplémentaires pour une plainte
        self.state_budgets = Permission() # APPEND = débloquer un nouveau budget, MANAGE = gérer les budjets, EDIT = gérer les sommes pour chaque budjet, READ = accéder aux infos concernant les budgets
        self.votes = Permission() # APPEND = déclencher un vote, MANAGE = fermer un vote, EDIT = /, READ = lire les propriétés d'un vote avant sa fermeture

    def merge(self, permissions: dict[str, str] | typing.Self):
        if isinstance(permissions, PositionPermissions):
            permissions = permissions.__dict__

        for key, val in permissions.items():
            perm: Permission = self.__getattribute__(key)
            perm.load(val)


class Position:
    """
    Position légale d'une entité

    ## Attributs
    - name: `str`\n
        Titre de la position
    - id: `str`\n
        Identifiant de la position
    - permissions: `.PositionPermissions`\n
        Permissions accordées à l'utilisateur
    """

    def __init__(self, id: str = 'inconnu') -> None:
        self.name: str = "Inconnue"
        self.id = id
        self.permissions: PositionPermissions = PositionPermissions()

    def __repr__(self):
        return self.id

class Entity:
    """
    Classe de référence pour les entités

    ## Attributs
    - id: `NSID`\n
        Identifiant de l'entité
    - name: `str`\n
        Nom d'usage de l'entité
    - registerDate: `int`\n
        Date d'enregistrement de l'entité
    - position: `.Position`\n
        Position légale de l'entité
    - additional: `dict`\n
        Infos supplémentaires exploitables par les bots
    """

    def __init__(self, id: NSID) -> None:
        self._url = "" # URL de l'entité pour une requête GET

        self.id: NSID = NSID(id) # ID hexadécimal de l'entité
        self.name: str = "Entité Inconnue"
        self.registerDate: int = 0
        self.position: Position = Position()
        self.additional: dict = {}

    def set_name(self, new_name: str) -> None:
        if len(new_name) > 32:
            raise ValueError(f"Name length mustn't exceed 32 characters.")

        res = requests.post(f"{self._url}/rename?name={new_name}", headers = default_headers)

        if res.status_code == 200:
            self.name = new_name
        else:
            print(res.status_code)
            res.raise_for_status()

    def set_position(self, position: Position) -> None:
        res = requests.post(f"{self._url}/change_position?position={position.id}", headers = default_headers)

        if res.status_code == 200:
            self.position = position
        else:
            res.raise_for_status()

    def add_link(self, key: str, value: str | int) -> None:
        if isinstance(value, str):
            _class = "string"
        elif isinstance(value, int):
            _class = "integer"
        else:
            raise TypeError("Only strings and integers can be recorded as an additional link")

        params = {
            "link": key,
            "value": value,
            "type": _class
        }

        query = "&".join(f"{k}={ urllib.parse.quote(v) }" for k, v in params.items())

        res = requests.post(f"{self._url}/add_link?{query}", headers = default_headers)

        if res.status_code == 200:
            self.additional[key] = value
        else:
            print(res.text)
            res.raise_for_status()

    def unlink(self, key: str) -> None:
        res = requests.post(f"{self._url}/remove_link?link={urllib.parse.quote(key)}", headers = default_headers)

        if res.status_code == 200:
            del self.additional[key]
        else:
            res.raise_for_status()

class User(Entity):
    """
    Entité individuelle

    ## Attributs
    - Tous les attributs de la classe `.Entity`
    - xp: `int`\n
        Points d'expérience de l'entité
    - boosts: `dict[str, int]`\n
        Ensemble des boosts dont bénéficie l'entité 
    - votes: `list[NSID]`\n
        Liste des votes auxquels a participé l'entité
    """

    def __init__(self, id: NSID) -> None:
        super().__init__(NSID(id))

        self.xp: int = 0
        self.boosts: dict[str, int] = {}
        self.votes: list[NSID] = []

    def _load(self, _data: dict):
        self.xp = _data['xp']
        self.boosts = _data['boosts']

        self.votes = [ NSID(vote) for vote in _data['votes'] ]

    def get_level(self) -> None:
        i = 0
        while self.xp > int(round(25 * (i * 2.5) ** 2, -2)):
            i += 1

        return i

    def add_xp(self, amount: int) -> None:
        boost = 0 if 0 in self.boosts.values() or amount <= 0 else max(list(self.boosts.values()) + [ 1 ])
        res = requests.post(f"{self._url}/add_xp?amount={amount * boost}", headers = default_headers)

        if res.status_code == 200:
            self.xp += amount * boost
        else:
            res.raise_for_status()

    def edit_boost(self, name: str, multiplier: int = -1) -> None:
        res = requests.post(f"{self._url}/edit_boost?boost={name}&multiplier={multiplier}", headers = default_headers)

        if res.status_code == 200:
            if multiplier >= 0:
                self.boosts[name] = multiplier
            else:
                del self.boosts[name]
        else:
            res.raise_for_status()

class MemberPermissions:
    """
    Permissions d'un utilisateur à l'échelle d'un groupe
    """

    def __init__(self) -> None:
        self.manage_organization = False # Renommer l'organisation, changer le logo
        self.manage_shares = False # Revaloriser les actions
        self.manage_roles = False # Changer les rôles des membres
        self.manage_members = False # Virer quelqu'un d'une entreprise, l'y inviter

    def edit(self, **permissions: bool) -> None:
        for perm in permissions.values():
            self.__setattr__(*perm)

class GroupMember:
    """
    Membre au sein d'une entité collective

    ## Attributs
    - permission_level: `dict[str, int]`\n
        Niveau d'accréditation du membre (0 = salarié, 4 = administrateur)
    """

    def __init__(self, id: NSID) -> None:
        self.id = id
        self.permission_level: dict = { # Échelle de permissions selon le groupe de travail
            "general": 0 
        }

    def group_permissions(self, team: str = "general") -> MemberPermissions:
        p = MemberPermissions()
        team_perms = self.permission_level[team]

        if team_perms >= 1: # Responsable
            p.manage_members = True

        if team_perms >= 2: # Superviseur
            p.manage_roles = True

        if team_perms >= 3: # Chef d'équipe
            pass

        if team_perms >= 4: # Directeur
            p.manage_shares = True
            p.manage_organization = True

        return p

class GroupInvite:
    def __init__(self, id: NSID):
        self.id: NSID = id
        self.team: str = "general"
        self.level: str = 0
        self._expires: int = round(time.time()) + 604800

class Organization(Entity):
    """
    Entité collective

    ## Attributs
    - Tous les attributs de la classe `.Entity`
    - owner: `.Entity`\n
        Utilisateur ou entreprise propriétaire de l'entité collective
    - avatar: `bytes`\n
        Avatar/logo de l'entité collective
    - certifications: `dict[str, int]`\n
        Liste des certifications et de leur date d'ajout
    - members: `list[.GroupMember]`\n
        Liste des membres de l'entreprise
    - parts: `list[.Share]`\n
        Liste des actions émises par l'entreprise
    """

    def __init__(self, id: NSID) -> None:
        super().__init__(NSID(id))

        self.owner: Entity = User(NSID(0x0))
        self.avatar: bytes = utils.open_asset('default_avatar.png')

        self.certifications: dict = {}
        self.members: list[GroupMember] = []
        self.invites: dict[GroupInvite] = []

    def _load(self, _data: dict):
        res = requests.get(f"{self._url}/avatar")

        if res.status_code == 200:
            self.avatar = res.content
        else:
            warnings.warn(f"Failed to get avatar for {self.id}")

        for _member in _data['members']:
            member = GroupMember(_member['id'])
            member.permission_level = _member['level']

            self.members.append(member)

        self.certifications = _data['certifications']

    def add_certification(self, certification: str, __expires: int = 2419200) -> None:
        res = requests.post(f"{self._url}/add_certification?name={certification}&duration={__expires}", headers = default_headers)

        if res.status_code == 200:
            self.certifications[certification] = int(round(time.time()) + __expires)
        else:
            res.raise_for_status()

    def has_certification(self, certification: str) -> bool:
        return certification in self.certifications.keys()

    def remove_certification(self, certification: str) -> None:
        res = requests.post(f"{self._url}/remove_certification?name={certification}", headers = default_headers)

        if res.status_code == 200:
            del self.certifications[certification]
        else:
            res.raise_for_status()

    def invite_member(self, member: NSID, level: int = 0, team: str = "general") -> None:
        if not isinstance(member, NSID):
            raise TypeError("L'entrée membre doit être de type NSID")

        res = requests.post(f"{self._url}/invite_member?id={member}&level={level}&team={team}", headers = default_headers)

        if res.status_code == 200:
            invite = GroupInvite(member)
            invite.team = team
            invite.level = level

            self.invites.append(invite)
        else:
            res.raise_for_status()

    def remove_member(self, member: GroupMember) -> None:
        for _member in self.members:
            if _member.id == member.id:
                self.members.remove(_member)

    def remove(self, member: GroupMember) -> None:
        self.remove_member(member)

    def set_owner(self, member: User) -> None:
        self.owner = member

    def get_members_by_attr(self, attribute: str = "id") -> list[str]:
        return [ member.__getattribute__(attribute) for member in self.members ]

    def save_avatar(self, data: bytes = None):
        if not data:
            return

        self.avatar = data


from ..cls.base import *
from ..cls.entities import *

from ..cls import entities # Pour les default_headers

class EntityInstance(Instance):
    """
    Instance qui vous permettra d'interagir avec les profils des membres ainsi que les différents métiers et secteurs d'activité.

    ## Informations disponibles
    - Profil des membres et des entreprises: `.User | .Organization | .Entity`
    - Participation d'un membre à différent votes: `.User | .Organization | .Entity`
    - Appartenance et permissions d'un membre dans un groupe: `.GroupMember.MemberPermissions`
    - Position légale et permissions d'une entité: `.Position.Permissions`
    - Sanctions et modifications d'une entité: `.Action[ .AdminAction | .Sanction ]`
    """

    def __init__(self, url: str, token: str = None) -> None:
        super().__init__(url, token)

        entities.default_headers = self.default_headers

    """
    ---- ENTITÉS ----
    """

    def get_entity(self, id: NSID, _class: str = None) -> User | Organization | Entity:
        """
        Fonction permettant de récupérer le profil public d'une entité.\n

        ## Paramètres
        id: `NSID`
            ID héxadécimal de l'entité à récupérer
        _class: `str`
            Classe du modèle à prendre (`.User` ou `.Organization`)

        ## Renvoie
        - `.User` dans le cas où l'entité choisie est un membre
        - `.Organization` dans le cas où c'est un groupe
        - `.Entity` dans le cas où c'est indéterminé
        """

        id = NSID(id)

        if _class == "user":
            _data = self._get_by_ID('individuals', id)
        elif _class == "group":
            _data = self._get_by_ID('organizations', id)
        else:
            _data = self._get_by_ID('entities', id)

        if _data is None: # ID inexistant chez les entités
            return None
        elif "xp" in _data.keys():
            _data['_type'] = 'user'
        elif "members" in _data.keys():
            _data['_type'] = 'organization'
        else:
            _data['_type'] = 'entity'

        if _data['_type'] == 'user':
            entity = User(id)
            entity._url = f"{self.url}/model/individuals/{id}"

            entity._load(_data)
        elif _data['_type'] == 'organization':
            entity = Organization(id)
            entity._url = f"{self.url}/model/organizations/{id}"
            entity.owner = self.get_entity(_data["owner_id"])

            entity._load(_data)
        else: 
            entity = Entity(id)

        entity.name = _data['name']
        entity.position = self.get_position(_data['position']) # Métier si c'est un utilisateur, domaine professionnel si c'est un collectif
        entity.registerDate = _data['register_date']

        for  key, value in _data.get('additional', {}).items():
            if isinstance(value, str) and value.startswith('\n'):
                entity.additional[key] = int(value[1:])
            else:
                entity.additional[key] = value

        return entity

    def get_entity_groups(self, entity: User) -> list[Organization]:
        print(entity._url)
        res = requests.get(f"{entity._url}/groups", headers = self.default_headers)

        if res.status_code == 200:
            return res.json()
        else:
            res.raise_for_status()
            return []

    def save_entity(self, entity: Entity):
        """
        Fonction permettant de créer ou modifier une entité.

        ## Paramètres
        entity: `.Entity`\n
            L'entité à sauvegarder
        """

        entity.id = NSID(entity.id)

        _data = {
            'id': entity.id,
            'name': entity.name,
            'position': entity.position.id,
            'register_date': entity.registerDate,
            'additional': {},
        }

        for key, value in entity.additional.items():
            if isinstance(value, int) and len(str(int)) >= 15:
                _data['additional'][key] = '\n' + str(value)
            elif type(value) in (str, int):
                _data['additional'][key] = value

        if type(entity) == Organization:
            _data['owner_id'] = NSID(entity.owner.id) if entity.owner else NSID("0")
            _data['members'] = []
            _data['certifications'] = entity.certifications

            for member in entity.members:
                _member = {
                    'id': NSID(member.id),
                    'level': member.permission_level
                }

                _data['members'] += [_member]

            entity.save_avatar()
        elif type(entity) == User:
            _data['xp'] = entity.xp
            _data['boosts'] = entity.boosts
            _data['votes'] = [ NSID(vote) for vote in entity.votes]
        else:
            return

        self._put_in_db(
            f"/new_model/{'individuals' if isinstance(entity, User) else 'organizations'}?id={urllib.parse.quote(entity.id)}&name={urllib.parse.quote(entity.name)}",
            _data,
            headers = self.default_headers,
            use_PUT = True
        )

        entity._url = f"{self.url}/model/{'individuals' if isinstance(entity, User) else 'organizations'}/{entity.id}"


    def delete_entity(self, entity: Entity):
        """
        Fonction permettant de supprimer le profil d'une entité

        ## Paramètres
        entity: `.Entity`\n
            L'entité à supprimer
        """

        res = requests.post(f"{entity._url}/delete", headers = self.default_headers,)

        if res.status_code != 200:
            res.raise_for_status()

    def fetch_entities(self, **query: typing.Any) -> list[ Entity | User | Organization ]:
        """
        Récupère une liste d'entités en fonction d'une requête.

        ## Paramètres
        query: `**dict`\n
            La requête pour filtrer les entités.

        ## Renvoie
        - `list[.Entity | .User | .Organization]`
        """

        if "_type" in query.keys():
            if query["_type"] == "individual":
                del query["_type"]
                _res = self.fetch('individuals', **query)
            elif query["_type"] == "organization":
                del query["_type"]
                _res = self.fetch('organizations', **query)
            else:
                del query["_type"]
                _res = self.fetch('entities', **query)
        else:
            _res = self.fetch('entities', **query)

        return [ self.get_entity(NSID(entity['id'])) for entity in _res if entity is not None ]

    def get_position(self, id: str) -> Position:
        """
        Récupère une position légale (métier, domaine professionnel).

        ## Paramètres
        id: `str`\n
            ID de la position (SENSIBLE À LA CASSE !)

        ## Renvoie
        - `.Position`
        """

        _data = self._get_by_ID('positions', id)

        if _data is None:
            return None

        position = Position(id)
        position.name = _data['name']
        position.permissions.merge(_data['permissions'])

        return position